#
#  --------------------------------------------------------------------------
#   Gurux Ltd
#
#
#
#  Filename: $HeadURL$
#
#  Version: $Revision$,
#                   $Date$
#                   $Author$
#
#  Copyright (c) Gurux Ltd
#
# ---------------------------------------------------------------------------
#
#   DESCRIPTION
#
#  This file is a part of Gurux Device Framework.
#
#  Gurux Device Framework is Open Source software; you can redistribute it
#  and/or modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; version 2 of the License.
#  Gurux Device Framework is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#  See the GNU General Public License for more details.
#
#  More information of Gurux products: http://www.gurux.org
#
#  This code is licensed under the GNU General Public License v2.
#  Full text may be retrieved at http://www.gnu.org/licenses/gpl-2.0.txt
# ---------------------------------------------------------------------------
import math
from .GXDLMSObject import GXDLMSObject
from .IGXDLMSBase import IGXDLMSBase
from ..enums import ErrorCode
from ..internal._GXCommon import _GXCommon
from ..GXByteBuffer import GXByteBuffer
from ..GXDateTime import GXDateTime
from ..enums import ObjectType, DataType, Unit

# pylint: disable=too-many-instance-attributes
class GXDLMSExtendedRegister(GXDLMSObject, IGXDLMSBase):
    """
    Online help:
    http://www.gurux.fi/Gurux.DLMS.Objects.GXDLMSExtendedRegister
    """

    def __init__(self, ln=None, sn=0):
        """
        Constructor.

        ln : Logical Name of the object.
        sn : Short Name of the object.
        """
        GXDLMSObject.__init__(self, ObjectType.EXTENDED_REGISTER, ln, sn)
        self.value = None
        self.scaler = 1
        self.unit = Unit.NONE
        self.status = None
        self.captureTime = None

    def getUIDataType(self, index):
        if index == 5:
            return DataType.DATETIME
        return super(GXDLMSExtendedRegister, self).getUIDataType(index)

    def getValues(self):
        return [self.logicalName,
                self.value,
                [self.scaler, self.unit],
                self.status,
                self.captureTime]

    #
    #      Returns collection of attributes to read.  If attribute is static
    #      and
    #      already read or device is returned HW error it is not returned.
    #
    def getAttributeIndexToRead(self, all_):
        attributes = list()
        #  LN is static and read only once.
        if all_ or not self.logicalName:
            attributes.append(1)
        #  ScalerUnit
        if all_ or not self.isRead(3):
            attributes.append(3)
        #  Value
        if all_ or self.canRead(2):
            attributes.append(2)
        #  Status
        if all_ or self.canRead(4):
            attributes.append(4)
        #  CaptureTime
        if all_ or self.canRead(5):
            attributes.append(5)
        return attributes

    #
    #      Returns amount of attributes.
    #
    def getAttributeCount(self):
        return 5

    def getMethodCount(self):
        return 0

    def getDataType(self, index):
        if index == 1:
            ret = DataType.OCTET_STRING
        elif index == 2:
            ret = super(GXDLMSExtendedRegister, self).getDataType(index)
        elif index == 3:
            ret = DataType.ARRAY
        elif index == 4:
            ret = super(GXDLMSExtendedRegister, self).getDataType(index)
        elif index == 5:
            ret = DataType.OCTET_STRING
        else:
            raise ValueError("getDataType failed. Invalid attribute index.")
        return ret
    #
    #      Returns value of given attribute.
    #
    def getValue(self, settings, e):
        if e.index == 1:
            return _GXCommon.logicalNameToBytes(self.logicalName)
        if e.index == 2:
            return self.value
        if e.index == 3:
            data = GXByteBuffer()
            data.setUInt8(DataType.STRUCTURE)
            data.setUInt8(2)
            _GXCommon.setData(data, DataType.INT8, math.floor(math.log(self.scaler, 10)))
            _GXCommon.setData(data, DataType.ENUM, int(self.unit))
            return data.array()
        if e.index == 4:
            return self.status
        if e.index == 5:
            return self.captureTime
        e.error = ErrorCode.READ_WRITE_DENIED
        return None

    #
    #      Set value of given attribute.
    #
    def setValue(self, settings, e):
        #pylint: disable=broad-except
        if e.index == 1:
            self.logicalName = _GXCommon.toLogicalName(e.value)
        elif e.index == 2:
            if self.scaler != 1 and e.value:
                try:
                    if settings.isServer:
                        self.value = e.value
                    else:
                        self.value = e.value * self.scaler
                except Exception:
                    #  Sometimes scaler is set for wrong Object type.
                    self.value = e.value
            else:
                self.value = e.value
        elif e.index == 3:
            #  Set default values.
            if not e.value:
                self.scaler = 0
                self.unit = 0
            else:
                if not e.value:
                    self.scaler = 0
                    self.unit = 0
                else:
                    self.scaler = math.pow(10, e.value[0])
                    self.unit = e.value[1]
        elif e.index == 4:
            self.status = e.value
        elif e.index == 5:
            if e.value is None:
                self.captureTime = GXDateTime()
            else:
                if isinstance(e.value, bytearray):
                    self.captureTime = _GXCommon.changeType(e.value, DataType.DATETIME)
                else:
                    self.captureTime = e.value
        else:
            e.error = ErrorCode.READ_WRITE_DENIED

    def load(self, reader):
        self.unit = Unit(reader.readElementContentAsInt("Unit", 0))
        self.scaler = reader.readElementContentAsDouble("Scaler", 1)
        self.value = reader.readElementContentAsObject("Value", None, self, 2)
        self.status = reader.readElementContentAsObject("Status", None, self, 4)
        self.captureTime = reader.readElementContentAsDateTime("CaptureTime")

    def save(self, writer):
        writer.writeElementString("Unit", self.unit)
        writer.writeElementString("Scaler", self.scaler, 1)
        writer.writeElementObject("Value", self.value, self.getDataType(2), self.getUIDataType(2))
        writer.writeElementObject("Status", self.status, self.getDataType(4), self.getUIDataType(4))
        writer.writeElementString("CaptureTime", self.captureTime)



#
#  --------------------------------------------------------------------------
#   Gurux Ltd
#
#
#
#  Filename: $HeadURL$
#
#  Version: $Revision$,
#                   $Date$
#                   $Author$
#
#  Copyright (c) Gurux Ltd
#
# ---------------------------------------------------------------------------
#
#   DESCRIPTION
#
#  This file is a part of Gurux Device Framework.
#
#  Gurux Device Framework is Open Source software; you can redistribute it
#  and/or modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; version 2 of the License.
#  Gurux Device Framework is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#  See the GNU General Public License for more details.
#
#  More information of Gurux products: http://www.gurux.org
#
#  This code is licensed under the GNU General Public License v2.
#  Full text may be retrieved at http://www.gnu.org/licenses/gpl-2.0.txt
# ---------------------------------------------------------------------------
#
#  --------------------------------------------------------------------------
#
from .GXUInt8 import GXUInt8

# BitString class is used with Bit strings.
class GXBitString:

    #
    # Constructor.
    #
    # @param val
    # Bit string value.
    #
    def __init__(self, val, count=None):
        if isinstance(val, (GXUInt8)):
            val = GXBitString._toBitString(val, 8)
            if count:
                val = val[0:count]
        self.value = val

    #
    # Reserved for internal use.
    #
    @classmethod
    def _toBitString(cls, value, count):
        count2 = count
        sb = ""
        if count2 > 0:
            if count2 > 8:
                count2 = 8
            pos = 7
            while pos != 8 - count2 - 1:
                if (value & (1 << pos)) != 0:
                    sb += '1'
                else:
                    sb += '0'
                pos -= 1
        return sb
    #
    # Bit string value.
    #
    def getValue(self):
        return self.value

    #
    # @param val
    # Bit string value.
    #
    def setValue(self, val):
        self.value = val

    def __str__(self):
        return self.value

    #
    # Bit string value as byte.
    #
    def toByte(self):
        val = 0
        if self.value:
            index = 7
            for it in self.value:
                if it == '1':
                    val |= (1 << index)
                elif it != '0':
                    raise ValueError("Invalid parameter.")
                index = index - 1
        return val



#
#  --------------------------------------------------------------------------
#   Gurux Ltd
#
#
#
#  Filename: $HeadURL$
#
#  Version: $Revision$,
#                   $Date$
#                   $Author$
#
#  Copyright (c) Gurux Ltd
#
# ---------------------------------------------------------------------------
#
#   DESCRIPTION
#
#  This file is a part of Gurux Device Framework.
#
#  Gurux Device Framework is Open Source software; you can redistribute it
#  and/or modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; version 2 of the License.
#  Gurux Device Framework is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#  See the GNU General Public License for more details.
#
#  More information of Gurux products: http://www.gurux.org
#
#  This code is licensed under the GNU General Public License v2.
#  Full text may be retrieved at http://www.gnu.org/licenses/gpl-2.0.txt
# ---------------------------------------------------------------------------
from .AccountCreditStatus import AccountCreditStatus
from .AccountStatus import AccountStatus
from .AddressConfigMode import AddressConfigMode
from .AddressState import AddressState
from .ApplicationContextName import ApplicationContextName
from .AssociationStatus import AssociationStatus
from .AutoAnswerMode import AutoAnswerMode
from .AutoAnswerStatus import AutoAnswerStatus
from .AutoConnectMode import AutoConnectMode
from .BaudRate import BaudRate
from .CertificateEntity import CertificateEntity
from .CertificateIdentificationType import CertificateIdentificationType
from .CertificateType import CertificateType
from .ChargeType import ChargeType
from .ClockBase import ClockBase
from .ControlMode import ControlMode
from .ControlState import ControlState
from .CreditCollectionConfiguration import CreditCollectionConfiguration
from .CreditConfiguration import CreditConfiguration
from .CreditStatus import CreditStatus
from .CreditType import CreditType
from .Currency import Currency
from .GlobalKeyType import GlobalKeyType
from .GsmCircuitSwitchStatus import GsmCircuitSwitchStatus
from .GsmPacketSwitchStatus import GsmPacketSwitchStatus
from .GsmStatus import GsmStatus
from .IecTwistedPairSetupMode import IecTwistedPairSetupMode
from .ImageTransferredBlocksStatus import ImageTransferredBlocksStatus
from .ImageTransferStatus import ImageTransferStatus
from .Ip4SetupIpOptionType import Ip4SetupIpOptionType
from .LocalPortResponseTime import LocalPortResponseTime
from .MessageType import MessageType
from .OpticalProtocolMode import OpticalProtocolMode
from .PaymentMode import PaymentMode
from .PppAuthenticationType import PppAuthenticationType
from .PppSetupIPCPOptionType import PppSetupIPCPOptionType
from .PppSetupLcpOptionType import PppSetupLcpOptionType
from .ScriptActionType import ScriptActionType
from .SecurityPolicy import SecurityPolicy
from .SecurityPolicy0 import SecurityPolicy0
from .SecuritySuite import SecuritySuite
from .ServiceType import ServiceType
from .SingleActionScheduleType import SingleActionScheduleType
from .SortMethod import SortMethod
from .TokenDelivery import TokenDelivery
from .TokenStatusCode import TokenStatusCode
from .MBusEncryptionKeyStatus import MBusEncryptionKeyStatus
from .Weekdays import Weekdays
from .MacState import MacState



#
#  --------------------------------------------------------------------------
#   Gurux Ltd
#
#
#
#  Filename:        $HeadURL$
#
#  Version:         $Revision$,
#                   $Date$
#                   $Author$
#
#  Copyright (c) Gurux Ltd
#
# ---------------------------------------------------------------------------
#
#   DESCRIPTION
#
#  This file is a part of Gurux Device Framework.
#
#  Gurux Device Framework is Open Source software; you can redistribute it
#  and/or modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; version 2 of the License.
#  Gurux Device Framework is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#  See the GNU General Public License for more details.
#
#  More information of Gurux products: http://www.gurux.org
#
#  This code is licensed under the GNU General Public License v2.
#  Full text may be retrieved at http://www.gnu.org/licenses/gpl-2.0.txt
# ---------------------------------------------------------------------------
import sys

#pylint: disable=no-name-in-module
if sys.version_info < (3, 6):
    __base = object
else:
    from enum import IntFlag
    __base = IntFlag

class Conformance(__base):
    """Enumerates all conformance bits.
    Client is using conformance to tell what kind of services it's interested
    and server is using conformance to tell what kind of services it can offer.

    All services enumerated here are not necessary supported by the client or the
    server.
    https://www.gurux.fi/Gurux.DLMS.Conformance
    """
    #pylint: disable=too-few-public-methods

    #Reserved zero conformance bit.
    #Not used at the moment.
    RESERVED_ZERO = 0x800000

    #Is General Protection supported by the meter or is client interested from
    #General Protection.
    #General Protection is used to secure connection between client and the
    #meter using Symmetric or ASymmetric ciphering or General Signing.
    GENERAL_PROTECTION = 0x400000

    #General Block Transfer mechanism is used to transport data that size is
    #longer than PDU size.
    GENERAL_BLOCK_TRANSFER = 0x200000

    #Is Read supported by the meter or is client interested from Read.
    #Read is used with Short Name Referencing to read value from the meter.
    READ = 0x100000

    #Is Write supported by the meter or is client interested from write data
    #to the meter.
    #Write is used with Short Name Referencing to write value to the meter.
    WRITE = 0x80000

    #Is Unconfirmed Write supported by the meter or is client interested from
    #write unconfirmed data to the meter.
    #Unconfirmed Write is used with Short Name Referencing to write value
    #without confirmation from success to the meter.
    UN_CONFIRMED_WRITE = 0x40000

    #Reserved six conformance bit.
    #Not used at the moment.
    RESERVED_SIX = 0x20000

    #Reserved seven conformance bit.
    #Not used at the moment.
    RESERVED_SEVEN = 0x10000

    #Is Attribute Set supported by the meter or is client interested from
    #write data with Attribute 0 to the meter.
    #Attribute 0 Set means that all COSEM object's attributes can be write
    #with one message.
    ATTRIBUTE_0_SUPPORTED_WITH_SET = 0x8000

    #Is Priority Management supported by the meter or is client interested
    #from Priority Management.
    #Priority Management is used by Logical Name referencing to handle urgent
    #messages.
    PRIORITY_MGMT_SUPPORTED = 0x4000

    #Is Attribute get supported by the meter or is client interested from read
    #data with Attribute 0 from the meter.
    #Attribute 0 Get means that all COSEM object's attributes can be read with
    #one message.
    ATTRIBUTE_0_SUPPORTED_WITH_GET = 0x2000

    #Is Block transfer supported by the meter or is client interested from
    #read data with blocks from the meter.
    #Read Block transfer is used to read data from the meter that is not fit
    #to one PDU.  Example reading historical data might take more than one PDU
    #and then block transfer is used.
    BLOCK_TRANSFER_WITH_GET_OR_READ = 0x1000

    #Is Block transfer supported by the meter or is client interested from
    #writing data with blocks to the meter.
    #Write Block transfer is used to write data from the meter that is not fit
    #to one PDU.  Example updating image =Firmware) might take more than one
    #PDU and then block transfer is used.
    BLOCK_TRANSFER_WITH_SET_OR_WRITE = 0x800

    #Is Block with action transfer supported by the meter or is client
    #interested from sending action data with blocks to the meter.
    #Action Block transfer is used to send action data from the meter that is
    #not fit to one PDU.  Example Public Key transfer might take more than one
    #PDU and then block transfer with action is used.
    BLOCK_TRANSFER_WITH_ACTION = 0x400

    #Is multiple references supported by the meter or is client interested
    #from multiple references.
    #Multiple references is used when reading or writing several object with
    #one message.  Example ReadByList is using it.
    MULTIPLE_REFERENCES = 0x200

    #Is Information reports supported by the meter.
    #Information reports is used with Short Name Referencing
    #
    #@see DATA_NOTIFICATION
    #@see EVENT_NOTIFICATION
    INFORMATION_REPORT = 0x100

    #Is Data Notification supported by the meter.
    #
    #@see INFORMATION_REPORT
    #@see EVENT_NOTIFICATION
    DATA_NOTIFICATION = 0x80

    #Is Access service supported by the meter or is client interested from
    #Access service.
    #Using Access service client can make several Read, Write or Action
    #requests with one command.
    ACCESS = 0x40

    #Is parameterized access supported by the meter or is client interested
    #from parameterized access.
    #Parameterized access is used with Short Name Referencing example if
    #Profile Generic is read by range or entry.
    #
    #@see SELECTIVE_ACCESS
    PARAMETERIZED_ACCESS = 0x20

    #Is Get supported by the meter or is client interested from Get.
    #Get is used with Logical Name Referencing to read value from the meter.
    GET = 0x10

    #Is Set supported by the meter or is client interested from Set.
    #Set is used with Logical Name Referencing to write value to the meter.
    SET = 0x8

    #Is selective access supported by the meter or is client interested from
    #selective access.
    #Selective access is used with Logical Name Referencing example if Profile
    #Generic is read by range or entry.
    #
    #@see PARAMETERIZED_ACCESS
    SELECTIVE_ACCESS = 0x4

    #Is Event Notification supported by the meter.
    #Meter is using Event Notifications with Logical Name referencing to send
    #events for given target.  Example if power down occurred.  Note!  Client
    #do
    #not need to have connection to the server when event notification is
    #send.
    #
    EVENT_NOTIFICATION = 0x2

    #Is actions supported by the meter or is client interested from actions.
    #
    #Actions are used example to reset historical data.
    ACTION = 0x1

    #None conformace is reserved for internal use.
    NONE = 0



#
#  --------------------------------------------------------------------------
#   Gurux Ltd
#
#
#
#  Filename: $HeadURL$
#
#  Version: $Revision$,
#                   $Date$
#                   $Author$
#
#  Copyright (c) Gurux Ltd
#
# ---------------------------------------------------------------------------
#
#   DESCRIPTION
#
#  This file is a part of Gurux Device Framework.
#
#  Gurux Device Framework is Open Source software; you can redistribute it
#  and/or modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; version 2 of the License.
#  Gurux Device Framework is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#  See the GNU General Public License for more details.
#
#  More information of Gurux products: http://www.gurux.org
#
#  This code is licensed under the GNU General Public License v2.
#  Full text may be retrieved at http://www.gnu.org/licenses/gpl-2.0.txt
# ---------------------------------------------------------------------------
import xml.etree.cElementTree as ET
try:
    from enum import Enum, IntEnum
except Exception:
    pass
from ..GXByteBuffer import GXByteBuffer
from ..GXDateTime import GXDateTime
from ..GXDLMSConverter import GXDLMSConverter
from ..internal._GXCommon import _GXCommon
from ..enums import DataType
from ..GXArray import GXArray
from ..GXStructure import GXStructure

###Python 2 requires this
#pylint: disable=bad-option-value,old-style-class
class GXXmlWriter:
    """
    Save COSEM object to the file.
    """

    def __init__(self):
        """
        Constructor.
        """
        self.objects = list()
        self.skipDefaults = False

    def getTarget(self):
        return self.objects[len(self.objects) - 1]

    # pylint: disable=unused-argument
    def writeStartElement(self, elementName, attributeName=None, value=None, newLine=True):
        target = None
        if value:
            target = ET.SubElement(self.getTarget(), elementName)
        else:
            target = ET.SubElement(self.getTarget(), elementName)

        if attributeName:
            target.set(attributeName, value)
        self.objects.append(target)
        return target

    def writeEndElement(self):
        self.objects.pop()


    def writeElementString(self, name, value, defaultValue=None):
        if not(value and self.skipDefaults) or value != defaultValue:
            if value is None:
                ET.SubElement(self.getTarget(), name)
            elif isinstance(value, str):
                ET.SubElement(self.getTarget(), name).text = value
            elif isinstance(value, GXDateTime):
                ET.SubElement(self.getTarget(), name).text = value.toFormatString()
            elif isinstance(value, bool):
                if value:
                    ET.SubElement(self.getTarget(), name).text = "1"
                else:
                    ET.SubElement(self.getTarget(), name).text = "0"
            elif isinstance(value, int):
                ET.SubElement(self.getTarget(), name).text = str(value)
            elif isinstance(value, (bytearray, bytes)):
                ET.SubElement(self.getTarget(), name).text = GXByteBuffer.hex(value)
            elif isinstance(value, (float)):
                ET.SubElement(self.getTarget(), name).text = str(value).replace(",", ".")

    def writeArray(self, data):
        if isinstance(data, (list)):
            arr = data
            for tmp in arr:
                if isinstance(tmp, bytearray):
                    self.writeElementObject("Item", tmp)
                elif isinstance(tmp, (GXArray,)):
                    self.writeStartElement("Item", "Type", str(int(DataType.ARRAY)), True)
                    self.writeArray(tmp)
                    self.writeEndElement()
                elif isinstance(tmp, (GXStructure,)):
                    self.writeStartElement("Item", "Type", str(int(DataType.STRUCTURE)), True)
                    self.writeArray(tmp)
                    self.writeEndElement()
                else:
                    self.writeElementObject("Item", tmp)

    #
    # Write object value to file.
    #
    # @param name
    # Object name.
    # @param value
    # Object value.
    #
    # pylint: disable=too-many-arguments
    def writeElementObject(self, name, value, dt=DataType.NONE, uiType=DataType.NONE):
        isEnum = False
        try:
            if isinstance(value, (Enum, IntEnum)):
                isEnum = True
        except Exception:
            pass
        if isEnum:
            raise ValueError("Datatype is enum.")

        if value or not self.skipDefaults:
            if value is None:
                target = self.writeStartElement(name)
                self.writeEndElement()
                return
            if dt == DataType.OCTET_STRING:
                if uiType == DataType.STRING:
                    value = str(value)
                elif uiType == DataType.OCTET_STRING:
                    value = GXByteBuffer.hex(value)
            elif dt != DataType.NONE and not isinstance(value, (float, GXDateTime)):
                value = GXDLMSConverter.changeType(value, dt)

            if dt == DataType.NONE:
                dt = _GXCommon.getDLMSDataType(value)

            target = self.writeStartElement(name, "Type", str(int(dt)), False)
            if uiType != DataType.NONE and uiType != dt and (uiType != DataType.STRING or dt == DataType.OCTET_STRING):
                target.set("UIType", str(int(uiType)))
            if dt in (DataType.ARRAY, DataType.STRUCTURE):
                self.writeArray(value)
            else:
                if isinstance(value, (float)):
                    target.set("UIType", str(int(DataType.FLOAT64)))
                if isinstance(value, GXDateTime):
                    target.text = value.toFormatString()
                elif isinstance(value, (bytearray, bytes)):
                    target.text = GXByteBuffer.hex(value)
                elif isinstance(value, bool):
                    if value:
                        target.text = "1"
                    else:
                        target.text = "0"
                else:
                    target.text = str(value)
            self.writeEndElement()



#
#  --------------------------------------------------------------------------
#   Gurux Ltd
#
#
#
#  Filename: $HeadURL$
#
#  Version: $Revision$,
#                   $Date$
#                   $Author$
#
#  Copyright (c) Gurux Ltd
#
# ---------------------------------------------------------------------------
#
#   DESCRIPTION
#
#  This file is a part of Gurux Device Framework.
#
#  Gurux Device Framework is Open Source software; you can redistribute it
#  and/or modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; version 2 of the License.
#  Gurux Device Framework is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#  See the GNU General Public License for more details.
#
#  More information of Gurux products: http://www.gurux.org
#
#  This code is licensed under the GNU General Public License v2.
#  Full text may be retrieved at http://www.gnu.org/licenses/gpl-2.0.txt
# ---------------------------------------------------------------------------
from .GXDLMSObject import GXDLMSObject
from .IGXDLMSBase import IGXDLMSBase
from ..enums import ErrorCode
from ..internal._GXCommon import _GXCommon
from ..GXByteBuffer import GXByteBuffer
from ..GXDateTime import GXDateTime
from ..enums import ObjectType, DataType
from .enums import AutoAnswerMode, AutoAnswerStatus

# pylint: disable=too-many-instance-attributes
class GXDLMSAutoAnswer(GXDLMSObject, IGXDLMSBase):
    """
    Online help:
    http://www.gurux.fi/Gurux.DLMS.Objects.GXDLMSAutoAnswer
    """

    def __init__(self, ln="0.0.2.2.0.255", sn=0):
        """
        Constructor.

        ln : Logical Name of the object.
        sn : Short Name of the object.
        """
        GXDLMSObject.__init__(self, ObjectType.AUTO_ANSWER, ln, sn)
        self.listeningWindow = list()
        self.mode = AutoAnswerMode.NONE
        self.status = AutoAnswerStatus.INACTIVE
        self.numberOfCalls = 0
        self.numberOfRingsInListeningWindow = 0
        self.numberOfRingsOutListeningWindow = 0

    def getValues(self):
        return [self.logicalName,
                self.mode,
                self.listeningWindow,
                self.status,
                self.numberOfCalls,
                [self.numberOfRingsInListeningWindow,
                 self.numberOfRingsOutListeningWindow]]
    #
    # Returns collection of attributes to read.  If attribute is static
    #      and
    # already read or device is returned HW error it is not returned.
    #
    def getAttributeIndexToRead(self, all_):
        attributes = list()
        #  LN is static and read only once.
        if all_ or not self.logicalName:
            attributes.append(1)
        #  Mode is static and read only once.
        if all_ or not self.isRead(2):
            attributes.append(2)
        #  ListeningWindow is static and read only once.
        if all_ or not self.isRead(3):
            attributes.append(3)
        #  Status is not static.
        if all_ or self.canRead(4):
            attributes.append(4)
        #  NumberOfCalls is static and read only once.
        if all_ or not self.isRead(5):
            attributes.append(5)
        #  NumberOfRingsInListeningWindow is static and read only once.
        if all_ or not self.isRead(6):
            attributes.append(6)
        return attributes

    #
    # Returns amount of attributes.
    #
    def getAttributeCount(self):
        return 6

    #
    # Returns amount of methods.
    #
    def getMethodCount(self):
        return 0

    def getDataType(self, index):
        if index == 1:
            ret = DataType.OCTET_STRING
        elif index == 2:
            ret = DataType.ENUM
        elif index == 3:
            ret = DataType.ARRAY
        elif index == 4:
            ret = DataType.ENUM
        elif index == 5:
            ret = DataType.UINT8
        elif index == 6:
            ret = DataType.ARRAY
        else:
            raise ValueError("getDataType failed. Invalid attribute index.")
        return ret
    #
    # Returns value of given attribute.
    #
    def getValue(self, settings, e):
        #pylint: disable=bad-option-value,redefined-variable-type
        if e.index == 1:
            ret = _GXCommon.logicalNameToBytes(self.logicalName)
        elif e.index == 2:
            ret = self.mode
        elif e.index == 3:
            cnt = len(self.listeningWindow)
            buff = GXByteBuffer()
            buff.setUInt8(DataType.ARRAY)
            #  Add count
            _GXCommon.setObjectCount(cnt, buff)
            if cnt != 0:
                for it in self.listeningWindow:
                    buff.setUInt8(DataType.STRUCTURE)
                    #  Count
                    buff.setUInt8(2)
                    #  Start time
                    _GXCommon.setData(buff, DataType.OCTET_STRING, it.getKey())
                    #  End time
                    _GXCommon.setData(buff, DataType.OCTET_STRING, it.value)
            ret = buff.array()
        elif e.index == 4:
            ret = self.status
        elif e.index == 5:
            ret = self.numberOfCalls
        elif e.index == 6:
            buff = GXByteBuffer()
            buff.setUInt8(DataType.STRUCTURE)
            _GXCommon.setObjectCount(2, buff)
            _GXCommon.setData(buff, DataType.UINT8, self.numberOfRingsInListeningWindow)
            _GXCommon.setData(buff, DataType.UINT8, self.numberOfRingsOutListeningWindow)
            ret = buff.array()
        else:
            e.error = ErrorCode.READ_WRITE_DENIED
        return ret

    #
    # Set value of given attribute.
    #
    def setValue(self, settings, e):
        if e.index == 1:
            self.logicalName = _GXCommon.toLogicalName(e.value)
        elif e.index == 2:
            self.mode = e.value
        elif e.index == 3:
            self.listeningWindow = []
            if e.value:
                for item in e.value:
                    start = _GXCommon.changeType(item[0], DataType.DATETIME)
                    end = _GXCommon.changeType(item[1], DataType.DATETIME)
                    self.listeningWindow.append((start, end))
        elif e.index == 4:
            self.status = e.value
        elif e.index == 5:
            self.numberOfCalls = e.value
        elif e.index == 6:
            self.numberOfRingsInListeningWindow = 0
            self.numberOfRingsOutListeningWindow = 0
            if e.value:
                self.numberOfRingsInListeningWindow = e.value[0]
                self.numberOfRingsOutListeningWindow = e.value[1]
        else:
            e.error = ErrorCode.READ_WRITE_DENIED

    def load(self, reader):
        self.mode = reader.readElementContentAsInt("Mode")
        self.listeningWindow = []
        if reader.isStartElement("ListeningWindow", True):
            while reader.isStartElement("Item", True):
                start = GXDateTime(reader.readElementContentAsString("Start"))
                end = GXDateTime(reader.readElementContentAsString("End"))
                self.listeningWindow.append((start, end))
            reader.readEndElement("ListeningWindow")
        self.status = reader.readElementContentAsInt("Status")
        self.numberOfCalls = reader.readElementContentAsInt("NumberOfCalls")
        self.numberOfRingsInListeningWindow = reader.readElementContentAsInt("NumberOfRingsInListeningWindow")
        self.numberOfRingsOutListeningWindow = reader.readElementContentAsInt("NumberOfRingsOutListeningWindow")

    def save(self, writer):
        writer.writeElementString("Mode", int(self.mode))
        writer.writeStartElement("ListeningWindow")
        if self.listeningWindow:
            for k, v in self.listeningWindow:
                writer.writeStartElement("Item")
                writer.writeElementString("Start", k.toFormatString())
                writer.writeElementString("End", v.toFormatString())
                writer.writeEndElement()
        writer.writeEndElement()
        writer.writeElementString("Status", int(self.status))
        writer.writeElementString("NumberOfCalls", self.numberOfCalls)
        writer.writeElementString("NumberOfRingsInListeningWindow", self.numberOfRingsInListeningWindow)
        writer.writeElementString("NumberOfRingsOutListeningWindow", self.numberOfRingsOutListeningWindow)



#
#  --------------------------------------------------------------------------
#   Gurux Ltd
#
#
#
#  Filename:        $HeadURL$
#
#  Version:         $Revision$,
#                   $Date$
#                   $Author$
#
#  Copyright (c) Gurux Ltd
#
# ---------------------------------------------------------------------------
#
#   DESCRIPTION
#
#  This file is a part of Gurux Device Framework.
#
#  Gurux Device Framework is Open Source software; you can redistribute it
#  and/or modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; version 2 of the License.
#  Gurux Device Framework is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#  See the GNU General Public License for more details.
#
#  More information of Gurux products: http://www.gurux.org
#
#  This code is licensed under the GNU General Public License v2.
#  Full text may be retrieved at http://www.gnu.org/licenses/gpl-2.0.txt
# ---------------------------------------------------------------------------
import sys

if sys.version_info < (3, 0):
    __base = object
else:
    from enum import IntEnum
    __base = IntEnum

class Command(__base):
    """DLMS commands."""

    #
    # No command to execute.
    #
    NONE = 0

    #
    # Initiate request.
    #
    INITIATE_REQUEST = 0x1

    #
    # Initiate response.
    #
    INITIATE_RESPONSE = 0x8

    #
    # Read request.
    #
    READ_REQUEST = 0x5

    #
    # Read response.
    #
    READ_RESPONSE = 0xC

    #
    # Write request.
    #
    WRITE_REQUEST = 0x6

    #
    # Write response.
    #
    WRITE_RESPONSE = 0xD

    #
    # Get request.
    #
    GET_REQUEST = 0xC0

    #
    # Get response.
    #
    GET_RESPONSE = 0xC4

    #
    # Set request.
    #
    SET_REQUEST = 0xC1

    #
    # Set response.
    #
    SET_RESPONSE = 0xC5

    #
    # Action request.
    #
    METHOD_REQUEST = 0xC3

    #
    # Action response.
    #
    METHOD_RESPONSE = 0xC7

    #
    # HDLC Disconnect Mode.
    #
    DISCONNECT_MODE = 0x1F

    #
    # HDLC Unacceptable frame.
    #
    UNACCEPTABLE_FRAME = 0x97

    #
    # HDLC SNRM request.
    #
    SNRM = 0x93

    #
    # HDLC UA request.
    #
    UA = 0x73

    #
    # AARQ request.
    #
    AARQ = 0x60

    #
    # AARE request.
    #
    AARE = 0x61

    #
    # Disconnect request for HDLC framing.
    #
    DISCONNECT_REQUEST = 0x53

    #
    # Release request.
    #
    RELEASE_REQUEST = 0x62

    #
    # Disconnect response.
    #
    RELEASE_RESPONSE = 0x63

    #
    # Confirmed Service Error.
    #
    CONFIRMED_SERVICE_ERROR = 0x0E

    #
    # Exception Response.
    #
    EXCEPTION_RESPONSE = 0xD8

    #
    # General Block Transfer.
    #
    GENERAL_BLOCK_TRANSFER = 0xE0

    #
    # Access Request.
    #
    ACCESS_REQUEST = 0xD9

    #
    # Access Response.
    #
    ACCESS_RESPONSE = 0xDA

    #
    # Data Notification request.
    #
    DATA_NOTIFICATION = 0x0F

    #
    # Glo get request.
    #
    GLO_GET_REQUEST = 0xC8

    #
    # Glo get response.
    #
    GLO_GET_RESPONSE = 0xCC

    #
    # Glo set request.
    #
    GLO_SET_REQUEST = 0xC9

    #
    # Glo set response.
    #
    GLO_SET_RESPONSE = 0xCD

    #
    # Glo event notification.
    #
    GLO_EVENT_NOTIFICATION = 0xCA

    #
    # Glo method request.
    #
    GLO_METHOD_REQUEST = 0xCB

    #
    # Glo method response.
    #
    GLO_METHOD_RESPONSE = 0xCF

    #
    # Glo Initiate request.
    #
    GLO_INITIATE_REQUEST = 0x21

    #
    # Glo read request.
    #
    GLO_READ_REQUEST = 37

    #
    # Glo write request.
    #
    GLO_WRITE_REQUEST = 38

    #
    # Glo Initiate response.
    #
    GLO_INITIATE_RESPONSE = 40

    #
    # Glo read response.
    #
    GLO_READ_RESPONSE = 44

    #
    # Glo write response.
    #
    GLO_WRITE_RESPONSE = 45

    #
    # General GLO ciphering.
    #
    GENERAL_GLO_CIPHERING = 0xDB

    #
    # General DED ciphering.
    #
    GENERAL_DED_CIPHERING = 0xDC

    #
    # General ciphering.
    #
    GENERAL_CIPHERING = 0xDD

    #
    # Information Report request.
    #
    INFORMATION_REPORT = 0x18

    #
    # Event Notification request.
    #
    EVENT_NOTIFICATION = 0xC2

    #
    # Ded initiate request.
    #
    DED_INITIATE_REQUEST = 65

    #
    # Ded read request.
    #
    DED_READ_REQUEST = 69

    #
    # Ded write request.
    #
    DED_WRITE_REQUEST = 70

    #
    # Ded initiate response.
    #
    DED_INITIATE_RESPONSE = 72

    #
    # Ded read response.
    #
    DED_READ_RESPONSE = 76

    #
    # Ded write response.
    #
    DED_WRITE_RESPONSE = 77

    #
    # Ded confirmed service error.
    #
    DED_CONFIRMED_SERVICE_ERROR = 78

    #
    # Ded confirmed write request.
    #
    DED_UNCONFIRMED_WRITE_REQUEST = 86

    #
    # Ded information report request.
    #
    DED_INFORMATION_REPORT_REQUEST = 88

    #
    # Ded get request.
    #
    DED_GET_REQUEST = 0xD0

    #
    # Ded get response.
    #
    DED_GET_RESPONSE = 0xD4

    #
    # Ded set request.
    #
    DED_SET_REQUEST = 0xD1

    #
    # Ded set response.
    #
    DED_SET_RESPONSE = 0xD5

    #
    # Ded event notification request.
    #
    DED_EVENT_NOTIFICATION = 0xD2

    #
    # Ded method request.
    #
    DED_METHOD_REQUEST = 0xD3

    #
    # Ded method response.
    #
    DED_METHOD_RESPONSE = 0xD7

    #
    # Request message from client to gateway.
    #
    GATEWAY_REQUEST = 0xE6

    #
    # Response message from gateway to client.
    #
    GATEWAY_RESPONSE = 0xE7

    @classmethod
    def toString(cls, value):
        str_ = None
        if value == Command.NONE:
            str_ = "None"
        elif value == Command.INITIATE_REQUEST:
            str_ = "InitiateRequest"
        elif value == Command.INITIATE_RESPONSE:
            str_ = "InitiateResponse"
        elif value == Command.READ_REQUEST:
            str_ = "ReadRequest"
        elif value == Command.READ_RESPONSE:
            str_ = "ReadResponse"
        elif value == Command.WRITE_REQUEST:
            str_ = "WriteRequest"
        elif value == Command.WRITE_RESPONSE:
            str_ = "WriteResponse"
        elif value == Command.GET_REQUEST:
            str_ = "GetRequest"
        elif value == Command.GET_RESPONSE:
            str_ = "GetResponse"
        elif value == Command.SET_REQUEST:
            str_ = "SetRequest"
        elif value == Command.SET_RESPONSE:
            str_ = "SetResponse"
        elif value == Command.METHOD_REQUEST:
            str_ = "MethodRequest"
        elif value == Command.METHOD_RESPONSE:
            str_ = "MethodResponse"
        elif value == Command.UNACCEPTABLE_FRAME:
            str_ = "UnacceptableFrame"
        elif value == Command.SNRM:
            str_ = "Snrm"
        elif value == Command.UA:
            str_ = "Ua"
        elif value == Command.AARQ:
            str_ = "Aarq"
        elif value == Command.AARE:
            str_ = "Aare"
        elif value == Command.DISCONNECT_REQUEST:
            str_ = "Disc"
        elif value == Command.RELEASE_REQUEST:
            str_ = "DisconnectRequest"
        elif value == Command.RELEASE_RESPONSE:
            str_ = "DisconnectResponse"
        elif value == Command.CONFIRMED_SERVICE_ERROR:
            str_ = "ConfirmedServiceError"
        elif value == Command.EXCEPTION_RESPONSE:
            str_ = "ExceptionResponse"
        elif value == Command.GENERAL_BLOCK_TRANSFER:
            str_ = "GeneralBlockTransfer"
        elif value == Command.ACCESS_REQUEST:
            str_ = "AccessRequest"
        elif value == Command.ACCESS_RESPONSE:
            str_ = "AccessResponse"
        elif value == Command.DATA_NOTIFICATION:
            str_ = "DataNotification"
        elif value == Command.GLO_GET_REQUEST:
            str_ = "GloGetRequest"
        elif value == Command.GLO_GET_RESPONSE:
            str_ = "GloGetResponse"
        elif value == Command.GLO_SET_REQUEST:
            str_ = "GloSetRequest"
        elif value == Command.GLO_SET_RESPONSE:
            str_ = "GloSetResponse"
        elif value == Command.GLO_EVENT_NOTIFICATION:
            str_ = "GloEventNotification"
        elif value == Command.GLO_METHOD_REQUEST:
            str_ = "GloMethodRequest"
        elif value == Command.GLO_METHOD_RESPONSE:
            str_ = "GloMethodResponse"
        elif value == Command.GLO_INITIATE_REQUEST:
            str_ = "GloInitiateRequest"
        elif value == Command.GLO_READ_REQUEST:
            str_ = "GloReadRequest"
        elif value == Command.GLO_WRITE_REQUEST:
            str_ = "GloWriteRequest"
        elif value == Command.GLO_INITIATE_RESPONSE:
            str_ = "GloInitiateResponse"
        elif value == Command.GLO_READ_RESPONSE:
            str_ = "GloReadResponse"
        elif value == Command.GLO_WRITE_RESPONSE:
            str_ = "GloWriteResponse"
        elif value == Command.GENERAL_GLO_CIPHERING:
            str_ = "GeneralGloCiphering"
        elif value == Command.GENERAL_DED_CIPHERING:
            str_ = "GeneralDedCiphering"
        elif value == Command.GENERAL_CIPHERING:
            str_ = "GeneralCiphering"
        elif value == Command.INFORMATION_REPORT:
            str_ = "InformationReport"
        elif value == Command.EVENT_NOTIFICATION:
            str_ = "EventNotification"
        elif value == Command.DED_GET_REQUEST:
            str_ = "DedGetRequest"
        elif value == Command.DED_GET_RESPONSE:
            str_ = "DedGetResponse"
        elif value == Command.DED_SET_REQUEST:
            str_ = "DedSetRequest"
        elif value == Command.DED_SET_RESPONSE:
            str_ = "DedSetResponse"
        elif value == Command.DED_EVENT_NOTIFICATION:
            str_ = "DedEventNotification"
        elif value == Command.DED_METHOD_REQUEST:
            str_ = "DedMethodRequest"
        elif value == Command.GATEWAY_REQUEST:
            str_ = "GatewayRequest "
        elif value == Command.GATEWAY_RESPONSE:
            str_ = "GatewayResponse "
        else:
            raise ValueError(str(value))
        return str_

    @classmethod
    def value_of(cls, value):
        if "None".lower() == value.lower():
            ret = Command.NONE
        elif "InitiateRequest".lower() == value.lower():
            ret = Command.INITIATE_REQUEST
        elif "InitiateResponse".lower() == value.lower():
            ret = Command.INITIATE_RESPONSE
        elif "ReadRequest".lower() == value.lower():
            ret = Command.READ_REQUEST
        elif "ReadResponse".lower() == value.lower():
            ret = Command.READ_RESPONSE
        elif "WriteRequest".lower() == value.lower():
            ret = Command.WRITE_REQUEST
        elif "WriteRequest".lower() == value.lower():
            ret = Command.WRITE_RESPONSE
        elif "WriteResponse".lower() == value.lower():
            ret = Command.WRITE_RESPONSE
        elif "GetRequest".lower() == value.lower():
            ret = Command.GET_REQUEST
        elif "GetResponse".lower() == value.lower():
            ret = Command.GET_RESPONSE
        elif "SetRequest".lower() == value.lower():
            ret = Command.SET_REQUEST
        elif "SetResponse".lower() == value.lower():
            ret = Command.SET_RESPONSE
        elif "MethodRequest".lower() == value.lower():
            ret = Command.METHOD_REQUEST
        elif "MethodResponse".lower() == value.lower():
            ret = Command.METHOD_RESPONSE
        elif "UnacceptableFrame".lower() == value.lower():
            ret = Command.UNACCEPTABLE_FRAME
        elif "Snrm".lower() == value.lower():
            ret = Command.SNRM
        elif "Ua".lower() == value.lower():
            ret = Command.UA
        elif "Aarq".lower() == value.lower():
            ret = Command.AARQ
        elif "Aare".lower() == value.lower():
            ret = Command.AARE
        elif "Disc".lower() == value.lower():
            ret = Command.DISCONNECT_REQUEST
        elif "DisconnectRequest".lower() == value.lower():
            ret = Command.RELEASE_REQUEST
        elif "DisconnectResponse".lower() == value.lower():
            ret = Command.RELEASE_RESPONSE
        elif "ConfirmedServiceError".lower() == value.lower():
            ret = Command.CONFIRMED_SERVICE_ERROR
        elif "ExceptionResponse".lower() == value.lower():
            ret = Command.EXCEPTION_RESPONSE
        elif "GeneralBlockTransfer".lower() == value.lower():
            ret = Command.GENERAL_BLOCK_TRANSFER
        elif "AccessRequest".lower() == value.lower():
            ret = Command.ACCESS_REQUEST
        elif "AccessResponse".lower() == value.lower():
            ret = Command.ACCESS_RESPONSE
        elif "DataNotification".lower() == value.lower():
            ret = Command.DATA_NOTIFICATION
        elif "GloGetRequest".lower() == value.lower():
            ret = Command.GLO_GET_REQUEST
        elif "GloGetResponse".lower() == value.lower():
            ret = Command.GLO_GET_RESPONSE
        elif "GloSetRequest".lower() == value.lower():
            ret = Command.GLO_SET_REQUEST
        elif "GloSetResponse".lower() == value.lower():
            ret = Command.GLO_SET_RESPONSE
        elif "GloEventNotification".lower() == value.lower():
            ret = Command.GLO_EVENT_NOTIFICATION
        elif "GloMethodRequest".lower() == value.lower():
            ret = Command.GLO_METHOD_REQUEST
        elif "GloMethodResponse".lower() == value.lower():
            ret = Command.GLO_METHOD_RESPONSE
        elif "GloInitiateRequest".lower() == value.lower():
            ret = Command.GLO_INITIATE_REQUEST
        elif "GloReadRequest".lower() == value.lower():
            ret = Command.GLO_READ_REQUEST
        elif "GloWriteRequest".lower() == value.lower():
            ret = Command.GLO_WRITE_REQUEST
        elif "GloInitiateResponse".lower() == value.lower():
            ret = Command.GLO_INITIATE_RESPONSE
        elif "GloReadResponse".lower() == value.lower():
            ret = Command.GLO_READ_RESPONSE
        elif "GloWriteResponse".lower() == value.lower():
            ret = Command.GLO_WRITE_RESPONSE
        elif "GeneralGloCiphering".lower() == value.lower():
            ret = Command.GENERAL_GLO_CIPHERING
        elif "GeneralDedCiphering".lower() == value.lower():
            ret = Command.GENERAL_DED_CIPHERING
        elif "GeneralCiphering".lower() == value.lower():
            ret = Command.GENERAL_CIPHERING
        elif "InformationReport".lower() == value.lower():
            ret = Command.INFORMATION_REPORT
        elif "EventNotification".lower() == value.lower():
            ret = Command.EVENT_NOTIFICATION
        elif "GatewayRequest".lower() == value.lower():
            ret = Command.GATEWAY_REQUEST
        elif "GatewayResponse".lower() == value.lower():
            ret = Command.GATEWAY_RESPONSE
        else:
            raise ValueError(value)
        return ret



#
#  --------------------------------------------------------------------------
#   Gurux Ltd
#
#
#
#  Filename: $HeadURL$
#
#  Version: $Revision$,
#                   $Date$
#                   $Author$
#
#  Copyright (c) Gurux Ltd
#
# ---------------------------------------------------------------------------
#
#   DESCRIPTION
#
#  This file is a part of Gurux Device Framework.
#
#  Gurux Device Framework is Open Source software; you can redistribute it
#  and/or modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; version 2 of the License.
#  Gurux Device Framework is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#  See the GNU General Public License for more details.
#
#  More information of Gurux products: http://www.gurux.org
#
#  This code is licensed under the GNU General Public License v2.
#  Full text may be retrieved at http://www.gnu.org/licenses/gpl-2.0.txt
# ---------------------------------------------------------------------------
from .GXAdjacentCell import GXAdjacentCell
from .GXApplicationContextName import GXApplicationContextName
from .GXAuthenticationMechanismName import GXAuthenticationMechanismName
from .GXChargePerUnitScaling import GXChargePerUnitScaling
from .GXChargeTable import GXChargeTable
from .GXCommodity import GXCommodity
from .GXCreditChargeConfiguration import GXCreditChargeConfiguration
from .GXCurrency import GXCurrency
from .GXDLMSRegister import GXDLMSRegister
from .GXDLMSDemandRegister import GXDLMSDemandRegister
from .GXDLMSRegisterMonitor import GXDLMSRegisterMonitor
from .GXDLMSRegisterActivation import GXDLMSRegisterActivation
from .GXDLMSExtendedRegister import GXDLMSExtendedRegister
from .GXDLMSAccount import GXDLMSAccount
from .GXDLMSActionItem import GXDLMSActionItem
from .GXDLMSActionSchedule import GXDLMSActionSchedule
from .GXDLMSActionSet import GXDLMSActionSet
from .GXDLMSActivityCalendar import GXDLMSActivityCalendar
from .GXDLMSAssociationLogicalName import GXDLMSAssociationLogicalName
from .GXDLMSAssociationShortName import GXDLMSAssociationShortName
from .GXDLMSAutoAnswer import GXDLMSAutoAnswer
from .GXDLMSAutoConnect import GXDLMSAutoConnect
from .GXDLMSCaptureObject import GXDLMSCaptureObject
from .GXDLMSCertificateInfo import GXDLMSCertificateInfo
from .GXDLMSCharge import GXDLMSCharge
from .GXDLMSClock import GXDLMSClock
from .GXDLMSCredit import GXDLMSCredit
from .GXDLMSData import GXDLMSData
from .GXDLMSDayProfile import GXDLMSDayProfile
from .GXDLMSDayProfileAction import GXDLMSDayProfileAction
from .GXDLMSDisconnectControl import GXDLMSDisconnectControl
from .GXDLMSEmergencyProfile import GXDLMSEmergencyProfile
from .GXDLMSGprsSetup import GXDLMSGprsSetup
from .GXDLMSGSMCellInfo import GXDLMSGSMCellInfo
from .GXDLMSGSMDiagnostic import GXDLMSGSMDiagnostic
from .GXDLMSHdlcSetup import GXDLMSHdlcSetup
from .GXDLMSIECLocalPortSetup import GXDLMSIECLocalPortSetup
from .GXDLMSIecTwistedPairSetup import GXDLMSIecTwistedPairSetup
from .GXDLMSImageActivateInfo import GXDLMSImageActivateInfo
from .GXDLMSImageTransfer import GXDLMSImageTransfer
from .GXDLMSIp4Setup import GXDLMSIp4Setup
from .GXDLMSIp4SetupIpOption import GXDLMSIp4SetupIpOption
from .GXDLMSIp6Setup import GXDLMSIp6Setup
from .GXDLMSLimiter import GXDLMSLimiter
from .GXDLMSMacAddressSetup import GXDLMSMacAddressSetup
from .GXDLMSMBusClient import GXDLMSMBusClient
from .GXDLMSMBusMasterPortSetup import GXDLMSMBusMasterPortSetup
from .GXDLMSMBusSlavePortSetup import GXDLMSMBusSlavePortSetup
from .GXDLMSModemConfiguration import GXDLMSModemConfiguration
from .GXDLMSModemInitialisation import GXDLMSModemInitialisation
from .GXDLMSMonitoredValue import GXDLMSMonitoredValue
from .GXDLMSObject import GXDLMSObject
from .GXDLMSObjectCollection import GXDLMSObjectCollection
from .GXDLMSObjectDefinition import GXDLMSObjectDefinition
from .GXDLMSParameterMonitor import GXDLMSParameterMonitor
from .GXDLMSPppSetup import GXDLMSPppSetup
from .GXDLMSPppSetupIPCPOption import GXDLMSPppSetupIPCPOption
from .GXDLMSPppSetupLcpOption import GXDLMSPppSetupLcpOption
from .GXDLMSProfileGeneric import GXDLMSProfileGeneric
from .GXDLMSPushSetup import GXDLMSPushSetup
from .GXDLMSQualityOfService import GXDLMSQualityOfService
from .GXDLMSSapAssignment import GXDLMSSapAssignment
from .GXDLMSSchedule import GXDLMSSchedule
from .GXDLMSScheduleEntry import GXDLMSScheduleEntry
from .GXDLMSScript import GXDLMSScript
from .GXDLMSScriptAction import GXDLMSScriptAction
from .GXDLMSScriptTable import GXDLMSScriptTable
from .GXDLMSSeasonProfile import GXDLMSSeasonProfile
from .GXDLMSSecuritySetup import GXDLMSSecuritySetup
from .GXDLMSSpecialDay import GXDLMSSpecialDay
from .GXDLMSSpecialDaysTable import GXDLMSSpecialDaysTable
from .GXDLMSTarget import GXDLMSTarget
from .GXDLMSTcpUdpSetup import GXDLMSTcpUdpSetup
from .GXDLMSTokenGateway import GXDLMSTokenGateway
from .GXDLMSWeekProfile import GXDLMSWeekProfile
from .GXNeighborDiscoverySetup import GXNeighborDiscoverySetup
from .GXSendDestinationAndMethod import GXSendDestinationAndMethod
from .GXTokenGatewayConfiguration import GXTokenGatewayConfiguration
from .GXUnitCharge import GXUnitCharge
from .GXxDLMSContextType import GXxDLMSContextType
from .GXXmlReader import GXXmlReader
from .GXXmlWriter import GXXmlWriter
from .GXXmlWriterSettings import GXXmlWriterSettings
from .IGXDLMSBase import IGXDLMSBase
from .GXDLMSUtilityTables import GXDLMSUtilityTables
from .GXDLMSLlcSscsSetup import GXDLMSLlcSscsSetup
from .GXDLMSPrimeNbOfdmPlcPhysicalLayerCounters import GXDLMSPrimeNbOfdmPlcPhysicalLayerCounters
from .GXDLMSPrimeNbOfdmPlcMacSetup import GXDLMSPrimeNbOfdmPlcMacSetup
from .GXDLMSPrimeNbOfdmPlcMacFunctionalParameters import GXDLMSPrimeNbOfdmPlcMacFunctionalParameters
from .GXDLMSPrimeNbOfdmPlcMacCounters import GXDLMSPrimeNbOfdmPlcMacCounters
from .GXDLMSPrimeNbOfdmPlcMacNetworkAdministrationData import GXDLMSPrimeNbOfdmPlcMacNetworkAdministrationData
from .GXDLMSPrimeNbOfdmPlcApplicationsIdentification import GXDLMSPrimeNbOfdmPlcApplicationsIdentification
from .GXMacMulticastEntry import GXMacMulticastEntry
from .GXMacDirectTable import GXMacDirectTable
from .GXMacAvailableSwitch import GXMacAvailableSwitch
from .GXMacPhyCommunication import GXMacPhyCommunication



#
#  --------------------------------------------------------------------------
#   Gurux Ltd
#
#
#
#  Filename: $HeadURL$
#
#  Version: $Revision$,
#                   $Date$
#                   $Author$
#
#  Copyright (c) Gurux Ltd
#
# ---------------------------------------------------------------------------
#
#   DESCRIPTION
#
#  This file is a part of Gurux Device Framework.
#
#  Gurux Device Framework is Open Source software; you can redistribute it
#  and/or modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; version 2 of the License.
#  Gurux Device Framework is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#  See the GNU General Public License for more details.
#
#  More information of Gurux products: http://www.gurux.org
#
#  This code is licensed under the GNU General Public License v2.
#  Full text may be retrieved at http://www.gnu.org/licenses/gpl-2.0.txt
# ---------------------------------------------------------------------------
from ..enums import InterfaceType, Authentication
from ..GXDLMSClient import GXDLMSClient
from ..GXCiphering import GXCiphering
from ..GXDLMSChipperingStream import GXDLMSChipperingStream

#pylint: disable=too-many-arguments
class GXDLMSSecureClient(GXDLMSClient):
    """
    GXDLMSSecureClient implements secure client where all messages are secured
    using transport security.
    """

    #
    # Constructor.
    #      *
    # @param useLogicalNameReferencing
    #            Is Logical Name referencing used.
    # @param clientAddress
    #            Server address.
    # @param serverAddress
    #            Client address.
    # @param forAuthentication
    #            Authentication type.
    # @param password
    #            Password if authentication is used.
    # @param interfaceType
    #            Object type.
    #
    def __init__(self, useLogicalNameReferencing=False, clientAddress=16, serverAddress=1, forAuthentication=Authentication.NONE, password=None, interfaceType=InterfaceType.HDLC):
        GXDLMSClient.__init__(self, useLogicalNameReferencing, clientAddress, serverAddress, forAuthentication, password, interfaceType)
        # Ciphering settings.
        self.settings.cipher = GXCiphering("ABCDEFGH".encode())

    #
    # Encrypt data using Key Encrypting Key.
    #      *
    # @param kek
    #            Key Encrypting Key, also known as Master key.
    # @param data
    #            Data to encrypt.
    # @return Encrypt data.
    #
    @classmethod
    def encrypt(cls, kek, data):
        if len(kek) != 16:
            raise ValueError("Key Encrypting Key")
        if not data:
            raise ValueError("data")
        gcm = GXDLMSChipperingStream(None, True, kek, None, None, 0)
        return gcm.encryptAes(data)

    #
    # Decrypt data using Key Encrypting Key.
    #      *
    # @param kek
    #            Key Encrypting Key, also known as Master key.
    # @param data
    #            Data to decrypt.
    # @return Decrypted data.
    #
    @classmethod
    def decrypt(cls, kek, data):
        if len(kek) != 16:
            raise ValueError("Key Encrypting Key")
        if not data:
            raise ValueError("data")
        gcm = GXDLMSChipperingStream(None, False, kek, None, None, 0)
        return gcm.decryptAes(data)

    def __getSecuritySuite(self):
        return self.ciphering.securitySuite

    def __setSecuritySuite(self, value):
        self.ciphering.securitySuite = value

    # Used security suite.
    securitySuite = property(__getSecuritySuite, __setSecuritySuite)

    def __getServerSystemTitle(self):
        return self.ciphering.sourceSystemTitle

    def __setServerSystemTitle(self, value):
        self.ciphering.sourceSystemTitle = value

    # Server system title.
    serverSystemTitle = property(__getServerSystemTitle, __setServerSystemTitle)

    def __getCiphering(self):
        return self.settings.cipher

    # Ciphering.
    ciphering = property(__getCiphering, None)



#
#  --------------------------------------------------------------------------
#   Gurux Ltd
#
#
#
#  Filename: $HeadURL$
#
#  Version: $Revision$,
#                   $Date$
#                   $Author$
#
#  Copyright (c) Gurux Ltd
#
# ---------------------------------------------------------------------------
#
#   DESCRIPTION
#
#  This file is a part of Gurux Device Framework.
#
#  Gurux Device Framework is Open Source software; you can redistribute it
#  and/or modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; version 2 of the License.
#  Gurux Device Framework is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#  See the GNU General Public License for more details.
#
#  More information of Gurux products: http://www.gurux.org
#
#  This code is licensed under the GNU General Public License v2.
#  Full text may be retrieved at http://www.gnu.org/licenses/gpl-2.0.txt
# ---------------------------------------------------------------------------
from .enums.Command import Command
# pylint: disable=bad-option-value,old-style-class
class GXDLMSXmlPdu:
    #pylint: disable=too-many-instance-attributes
    def __init__(self, command=0, xml=None, pdu=None):
        # XML Node.
        self.xmlNode = xml
        # Command.
        self.command = command
        # Description of the test.
        self.description = None
        # Shown error if this test fails.
        self.error = None
        # Error url if test fails.
        self.errorUrl = None
        # Generated Pdu.
        self.data = pdu
        # Sleep time in milliseconds.
        self.privateSleep = 0

    @classmethod
    def getOuterXml(cls, node):
        return ""

    #
    # Return PDU as XML string.
    #
    def getPduAsXml(self):
        if self.xmlNode is None:
            return ""
        return self.getOuterXml(self.xmlNode)

    @classmethod
    def compare(cls, expectedNode, actualNode, list_):
        cnt = expectedNode.getChildNodes().getLength()
        if expectedNode.getNodeName().compareTo(actualNode.getNodeName()) != 0:
            a = expectedNode.getAttributes().getNamedItem("Value")
            if expectedNode.getNodeName().compareTo("None") == 0 and a and a.getNodeValue().compareTo("*") == 0:
                return
            list_.append(expectedNode.getNodeName() + "-" + actualNode.getNodeName())
            return
        if cnt != actualNode.getChildNodes().getLength():
            #  If we are reading array items count might vary.
            if expectedNode.getNodeName() == "Array" or expectedNode.getNodeName() == "Structure":
                #  Check only first If meter is returning more nodes what
                #  wehave in template.
                if not cnt < actualNode.getChildNodes().getLength():
                    cnt = actualNode.getChildNodes().getLength()
            else:
                list_.append("Different amount: " + expectedNode.getNodeName() + "-" + actualNode.getNodeName())
                return
        pos = 0
        while pos != cnt:
            if actualNode.getChildNodes().item(pos) is None:
                list_.append("Different values. Expected: '" + expectedNode.getChildNodes().item(pos).getNodeValue() + "'. Actual: 'null'.")
            elif actualNode.getChildNodes().item(pos).getChildNodes().getLength() != 0:
                cls.compare(expectedNode.getChildNodes().item(pos), actualNode.getChildNodes().item(pos), list_)
            elif expectedNode.getChildNodes().item(pos).getNodeValue().compareTo(actualNode.getChildNodes().item(pos).getNodeValue()) != 0:
                a = expectedNode.getChildNodes().item(pos).getAttributes().getNamedItem("Value")
                if a is None or (expectedNode.getChildNodes().item(pos).getNodeName().compareTo("None") != 0 and expectedNode.getChildNodes().item(pos).getNodeName().compareTo(actualNode.getChildNodes().item(pos).getNodeName()) != 0) or a.getNodeValue().compareTo("*") != 0:
                    if not expectedNode.getFirstChild().getNodeName() == "Structure" and not expectedNode.getFirstChild().getNodeName() == "Array" and not expectedNode.getParentNode().getNodeName() == "Array":
                        list_.append("Different values. Expected: '" + expectedNode.getChildNodes().item(pos).getNodeValue() + "'. Actual: '" + actualNode.getChildNodes().item(pos).getNodeValue() + "'.")
            pos += 1

    def isRequest(self):
        return self.command in (Command.SNRM, Command.AARQ, Command.READ_REQUEST, Command.GLO_READ_REQUEST,\
            Command.WRITE_REQUEST, Command.GLO_WRITE_REQUEST, Command.GET_REQUEST, Command.GLO_GET_REQUEST,\
            Command.SET_REQUEST, Command.GLO_SET_REQUEST, Command.METHOD_REQUEST, Command.GLO_METHOD_REQUEST,\
            Command.DISCONNECT_REQUEST, Command.RELEASE_REQUEST)

    def __str__(self):
        return self.getPduAsXml()



#
#  --------------------------------------------------------------------------
#   Gurux Ltd
#
#
#
#  Filename: $HeadURL$
#
#  Version: $Revision$,
#                   $Date$
#                   $Author$
#
#  Copyright (c) Gurux Ltd
#
# ---------------------------------------------------------------------------
#
#   DESCRIPTION
#
#  This file is a part of Gurux Device Framework.
#
#  Gurux Device Framework is Open Source software; you can redistribute it
#  and/or modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; version 2 of the License.
#  Gurux Device Framework is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#  See the GNU General Public License for more details.
#
#  More information of Gurux products: http://www.gurux.org
#
#  This code is licensed under the GNU General Public License v2.
#  Full text may be retrieved at http://www.gnu.org/licenses/gpl-2.0.txt
# ---------------------------------------------------------------------------
import sys

#pylint: disable=no-name-in-module
if sys.version_info < (3, 6):
    __base = object
else:
    from enum import IntFlag
    __base = IntFlag

class AccountCreditStatus(__base):
    """Enumerates account credit status modes.
    Online help:
    http://www.gurux.fi/Gurux.DLMS.Objects.GXDLMSAccount
    """
    #pylint: disable=too-few-public-methods

    #
    # In credit.
    #
    IN_CREDIT = 0x80

    #
    # Low credit.
    #
    LOW_CREDIT = 0x40

    #
    # Next credit enabled.
    #
    NEXT_CREDIT_ENABLED = 0x20

    #
    # Next credit selectable.
    #
    NEXT_CREDIT_SELECTABLE = 0x10

    #
    # Credit reference list.
    #
    CREDIT_REFERENCE_LIST = 0x8

    #
    # Selectable credit in use.
    #
    SELECTABLE_CREDIT_IN_USE = 0x4

    #
    # Out of credit.
    #
    OUT_OF_CREDIT = 0x2

    #
    # Reserved.
    #
    RESERVED = 0x1



#
#  --------------------------------------------------------------------------
#   Gurux Ltd
#
#
#
#  Filename: $HeadURL$
#
#  Version: $Revision$,
#                   $Date$
#                   $Author$
#
#  Copyright (c) Gurux Ltd
#
# ---------------------------------------------------------------------------
#
#   DESCRIPTION
#
#  This file is a part of Gurux Device Framework.
#
#  Gurux Device Framework is Open Source software; you can redistribute it
#  and/or modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; version 2 of the License.
#  Gurux Device Framework is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#  See the GNU General Public License for more details.
#
#  More information of Gurux products: http://www.gurux.org
#
#  This code is licensed under the GNU General Public License v2.
#  Full text may be retrieved at http://www.gnu.org/licenses/gpl-2.0.txt
# ---------------------------------------------------------------------------
from .GXDLMSObject import GXDLMSObject
from .IGXDLMSBase import IGXDLMSBase
from ..enums import ErrorCode
from ..internal._GXCommon import _GXCommon
from ..GXByteBuffer import GXByteBuffer
from ..GXDateTime import GXDateTime
from ..enums import ObjectType, DataType
from .enums import PaymentMode, AccountStatus, AccountCreditStatus
from .GXCurrency import GXCurrency
from .GXTokenGatewayConfiguration import GXTokenGatewayConfiguration
from .GXCreditChargeConfiguration import GXCreditChargeConfiguration

# pylint: disable=too-many-instance-attributes
class GXDLMSAccount(GXDLMSObject, IGXDLMSBase):

    """
    Online help:
    http://www.gurux.fi/Gurux.DLMS.Objects.GXDLMSAccount
    """
    def __init__(self, ln="0.0.19.0.0.255", sn=0):
        """
        Constructor.

        ln : Logical Name of the object.
        sn : Short Name of the object.
        """
        GXDLMSObject.__init__(self, ObjectType.ACCOUNT, ln, sn)
        self.paymentMode = PaymentMode.CREDIT
        self.accountStatus = AccountStatus.NEW_INACTIVE_ACCOUNT
        self.creditReferences = list()
        self.chargeReferences = list()
        self.creditChargeConfigurations = list()
        self.tokenGatewayConfigurations = list()
        self.currency = GXCurrency()
        self.currentCreditInUse = 0
        self.currentCreditStatus = AccountCreditStatus.IN_CREDIT
        self.availableCredit = 0
        self.amountToClear = 0
        self.clearanceThreshold = 0
        self.aggregatedDebt = 0
        self.accountActivationTime = None
        self.accountClosureTime = None
        self.lowCreditThreshold = 0
        self.nextCreditAvailableThreshold = 0
        self.maxProvision = 0
        self.maxProvisionPeriod = 0

    def getValues(self):
        return [self.logicalName,
                [self.paymentMode, self.accountStatus],
                self.currentCreditInUse,
                self.currentCreditStatus,
                self.availableCredit,
                self.amountToClear,
                self.clearanceThreshold,
                self.aggregatedDebt,
                self.creditReferences,
                self.chargeReferences,
                self.creditChargeConfigurations,
                self.tokenGatewayConfigurations,
                self.accountActivationTime,
                self.accountClosureTime,
                self.currency,
                self.lowCreditThreshold,
                self.nextCreditAvailableThreshold,
                self.maxProvision,
                self.maxProvisionPeriod]

    def activate(self, client):
        """Activate account."""
        return client.method(self.getName(), self.objectType, 1, 0, DataType.INT8)

    def close(self, client):
        """Close account."""
        return client.method(self.getName(), self.objectType, 2, 0, DataType.INT8)

    def reset(self, client):
        """Reset account."""
        return client.method(self.getName(), self.objectType, 3, 0, DataType.INT8)

    #
    # Returns collection of attributes to read.  If attribute is static
    #      and
    # already read or device is returned HW error it is not returned.
    #
    def getAttributeIndexToRead(self, all_):
        attributes = list()
        #  LN is static and read only once.
        if all_ or not self.logicalName:
            attributes.append(1)
        #  PaymentMode, AccountStatus
        if all_ or self.canRead(2):
            attributes.append(2)
        #  CurrentCreditInUse
        if all_ or self.canRead(3):
            attributes.append(3)
        #  CurrentCreditStatus
        if all_ or self.canRead(4):
            attributes.append(4)
        #  AvailableCredit
        if all_ or self.canRead(5):
            attributes.append(5)
        #  AmountToClear
        if all_ or self.canRead(6):
            attributes.append(6)
        #  ClearanceThreshold
        if all_ or self.canRead(7):
            attributes.append(7)
        #  AggregatedDebt
        if all_ or self.canRead(8):
            attributes.append(8)
        #  CreditReferences
        if all_ or self.canRead(9):
            attributes.append(9)
        #  ChargeReferences
        if all_ or self.canRead(10):
            attributes.append(10)
        #  CreditChargeConfigurations
        if all_ or self.canRead(11):
            attributes.append(11)
        #  TokenGatewayConfigurations
        if all_ or self.canRead(12):
            attributes.append(12)
        #  AccountActivationTime
        if all_ or self.canRead(13):
            attributes.append(13)
        #  AccountClosureTime
        if all_ or self.canRead(14):
            attributes.append(14)
        #  Currency
        if all_ or self.canRead(15):
            attributes.append(15)
        #  LowCreditThreshold
        if all_ or self.canRead(16):
            attributes.append(16)
        #  NextCreditAvailableThreshold
        if all_ or self.canRead(17):
            attributes.append(17)
        #  MaxProvision
        if all_ or self.canRead(18):
            attributes.append(18)
        #  MaxProvisionPeriod
        if all_ or self.canRead(19):
            attributes.append(19)
        return attributes

    #
    # Returns amount of attributes.
    #
    def getAttributeCount(self):
        return 19

    #
    # Returns amount of methods.
    #
    def getMethodCount(self):
        return 3

    def getDataType(self, index):
        if index == 1:
            ret = DataType.OCTET_STRING
        elif index == 2:
            ret = DataType.STRUCTURE
        elif index == 3:
            ret = DataType.UINT8
        elif index == 4:
            ret = DataType.BITSTRING
        elif index == 5:
            ret = DataType.INT32
        elif index == 6:
            ret = DataType.INT32
        elif index == 7:
            ret = DataType.INT32
        elif index == 8:
            ret = DataType.INT32
        elif index == 9:
            ret = DataType.ARRAY
        elif index == 10:
            ret = DataType.ARRAY
        elif index == 11:
            ret = DataType.ARRAY
        elif index == 12:
            ret = DataType.ARRAY
        elif index == 13:
            ret = DataType.OCTET_STRING
        elif index == 14:
            ret = DataType.OCTET_STRING
        elif index == 15:
            ret = DataType.STRUCTURE
        elif index == 16:
            ret = DataType.INT32
        elif index == 17:
            ret = DataType.INT32
        elif index == 18:
            ret = DataType.UINT16
        elif index == 19:
            ret = DataType.INT32
        else:
            raise ValueError("getDataType failed. Invalid attribute index.")
        return ret

    #
    # Returns value of given attribute.
    #
    def getValue(self, settings, e):
        #pylint: disable=bad-option-value,redefined-variable-type
        if e.index == 1:
            ret = _GXCommon.logicalNameToBytes(self.logicalName)
        elif e.index == 2:
            bb = GXByteBuffer()
            bb.setUInt8(DataType.STRUCTURE)
            bb.setUInt8(2)
            bb.setUInt8(DataType.ENUM)
            bb.setUInt8(self.accountStatus)
            bb.setUInt8(DataType.ENUM)
            bb.setUInt8(self.paymentMode)
            ret = bb.array()
        elif e.index == 3:
            ret = self.currentCreditInUse
        elif e.index == 4:
            ret = self.currentCreditStatus
        elif e.index == 5:
            ret = self.availableCredit
        elif e.index == 6:
            ret = self.amountToClear
        elif e.index == 7:
            ret = self.clearanceThreshold
        elif e.index == 8:
            ret = self.aggregatedDebt
        elif e.index == 9:
            bb = GXByteBuffer()
            bb.setUInt8(DataType.ARRAY)
            if not self.creditReferences:
                bb.setUInt8(0)
            else:
                _GXCommon.setObjectCount(len(self.creditReferences), bb)
                for it in self.creditReferences:
                    bb.setUInt8(DataType.OCTET_STRING)
                    bb.setUInt8(6)
                    bb.set(_GXCommon.logicalNameToBytes(it))
            ret = bb.array()
        elif e.index == 10:
            bb = GXByteBuffer()
            bb.setUInt8(DataType.ARRAY)
            if not self.chargeReferences:
                bb.setUInt8(0)
            else:
                _GXCommon.setObjectCount(len(self.chargeReferences), bb)
                for it in self.chargeReferences:
                    bb.setUInt8(DataType.OCTET_STRING)
                    bb.setUInt8(6)
                    bb.set(_GXCommon.logicalNameToBytes(it))
            ret = bb.array()
        elif e.index == 11:
            bb = GXByteBuffer()
            bb.setUInt8(DataType.ARRAY)
            if not self.creditChargeConfigurations:
                bb.setUInt8(0)
            else:
                _GXCommon.setObjectCount(len(self.creditChargeConfigurations), bb)
                for it in self.creditChargeConfigurations:
                    bb.setUInt8(DataType.STRUCTURE)
                    bb.setUInt8(3)
                    bb.setUInt8(DataType.OCTET_STRING)
                    bb.setUInt8(6)
                    bb.set(_GXCommon.logicalNameToBytes(it.creditReference))
                    bb.setUInt8(DataType.OCTET_STRING)
                    bb.setUInt8(6)
                    bb.set(_GXCommon.logicalNameToBytes(it.chargeReference))
                    _GXCommon.setData(bb, DataType.BITSTRING, it.collectionConfiguration)
            ret = bb.array()
        elif e.index == 12:
            bb = GXByteBuffer()
            bb.setUInt8(DataType.ARRAY)
            if not self.tokenGatewayConfigurations:
                bb.setUInt8(0)
            else:
                _GXCommon.setObjectCount(len(self.tokenGatewayConfigurations), bb)
                for it in self.tokenGatewayConfigurations:
                    bb.setUInt8(DataType.STRUCTURE)
                    bb.setUInt8(2)
                    bb.setUInt8(DataType.OCTET_STRING)
                    bb.setUInt8(6)
                    bb.set(_GXCommon.logicalNameToBytes(it.creditReference))
                    bb.setUInt8(DataType.UINT8)
                    bb.setUInt8(it.tokenProportion)
            ret = bb.array()
        elif e.index == 13:
            ret = self.accountActivationTime
        elif e.index == 14:
            ret = self.accountClosureTime
        elif e.index == 15:
            bb = GXByteBuffer()
            bb.setUInt8(DataType.STRUCTURE)
            bb.setUInt8(3)
            _GXCommon.setData(bb, DataType.STRING_UTF8, self.currency.name)
            _GXCommon.setData(bb, DataType.INT8, self.currency.scale)
            _GXCommon.setData(bb, DataType.ENUM, self.currency.unit)
            ret = bb.array()
        elif e.index == 16:
            ret = self.lowCreditThreshold
        elif e.index == 17:
            ret = self.nextCreditAvailableThreshold
        elif e.index == 18:
            ret = self.maxProvision
        elif e.index == 19:
            ret = self.maxProvisionPeriod
        else:
            e.error = ErrorCode.READ_WRITE_DENIED
        return ret

    #
    # Set value of given attribute.
    #
    def setValue(self, settings, e):
        if e.index == 1:
            self.logicalName = _GXCommon.toLogicalName(e.value)
        elif e.index == 2:
            self.accountStatus = e.value[0]
            self.paymentMode = e.value[1]
        elif e.index == 3:
            self.currentCreditInUse = e.value
        elif e.index == 4:
            bb = GXByteBuffer()
            _GXCommon.setBitString(bb, e.value, False)
            self.currentCreditStatus = bb.getUInt8(0)
        elif e.index == 5:
            self.availableCredit = e.value
        elif e.index == 6:
            self.amountToClear = e.value
        elif e.index == 7:
            self.clearanceThreshold = e.value
        elif e.index == 8:
            self.aggregatedDebt = e.value
        elif e.index == 9:
            self.creditReferences = []
            if e.value:
                for it in e.value:
                    self.creditReferences.append(_GXCommon.toLogicalName(it))
        elif e.index == 10:
            self.chargeReferences = []
            if e.value:
                for it in e.value:
                    self.chargeReferences.append(_GXCommon.toLogicalName(it))
        elif e.index == 11:
            #pylint: disable=bad-option-value,redefined-variable-type
            self.creditChargeConfigurations = []
            if e.value:
                for it in e.value:
                    item = GXCreditChargeConfiguration()
                    item.creditReference = _GXCommon.toLogicalName(it[0])
                    item.chargeReference = _GXCommon.toLogicalName(it[1])
                    bb = GXByteBuffer()
                    _GXCommon.setBitString(bb, it[2], False)
                    item.collectionConfiguration = bb.getUInt8(0)
                    self.creditChargeConfigurations.append(item)
        elif e.index == 12:
            #pylint: disable=bad-option-value,redefined-variable-type
            self.tokenGatewayConfigurations = []
            if e.value:
                for it in e.value:
                    item = GXTokenGatewayConfiguration()
                    item.creditReference = _GXCommon.toLogicalName(it[0])
                    item.tokenProportion = it[1]
                    self.tokenGatewayConfigurations.append(item)
        elif e.index == 13:
            if not e.value:
                self.accountActivationTime = GXDateTime()
            else:
                tmp = None
                if isinstance(e.value, bytearray):
                    tmp = _GXCommon.changeType(e.value, DataType.DATETIME)
                else:
                    tmp = e.value
                self.accountActivationTime = tmp
        elif e.index == 14:
            if not e.value:
                self.accountClosureTime = GXDateTime()
            else:
                tmp = None
                if isinstance(e.value, bytearray):
                    tmp = _GXCommon.changeType(e.value, DataType.DATETIME)
                else:
                    tmp = e.value
                self.accountClosureTime = tmp
        elif e.index == 15:
            tmp = e.value
            self.currency.name = str(tmp[0])
            self.currency.scale = tmp[1]
            self.currency.unit = tmp[2]
        elif e.index == 16:
            self.lowCreditThreshold = e.value
        elif e.index == 17:
            self.nextCreditAvailableThreshold = e.value
        elif e.index == 18:
            self.maxProvision = e.value
        elif e.index == 19:
            self.maxProvisionPeriod = e.value
        else:
            e.error = ErrorCode.READ_WRITE_DENIED

    @classmethod
    def loadReferences(cls, reader, name, list_):
        list_ = []
        if reader.isStartElement(name, True):
            while reader.isStartElement("Item", True):
                list_.append(reader.readElementContentAsString("Name"))
            reader.readEndElement(name)

    @classmethod
    def loadCreditChargeConfigurations(cls, reader, list_):
        list_ = []
        if reader.isStartElement("CreditChargeConfigurations", True):
            while reader.isStartElement("Item", True):
                it = GXCreditChargeConfiguration()
                it.creditReference = reader.readElementContentAsString("Credit")
                it.chargeReference = reader.readElementContentAsString("Charge")
                it.collectionConfiguration = reader.readElementContentAsInt("Configuration")
                list_.append(it)
            reader.readEndElement("CreditChargeConfigurations")

    @classmethod
    def loadTokenGatewayConfigurations(cls, reader, list_):
        list_ = []
        if reader.isStartElement("TokenGatewayConfigurations", True):
            while reader.isStartElement("Item", True):
                it = GXTokenGatewayConfiguration()
                it.creditReference = reader.readElementContentAsString("Credit")
                it.tokenProportion = reader.readElementContentAsInt("Token")
                list_.append(it)
            reader.readEndElement("TokenGatewayConfigurations")

    def load(self, reader):
        self.paymentMode = reader.readElementContentAsInt("PaymentMode")
        self.accountStatus = reader.readElementContentAsInt("AccountStatus")
        self.currentCreditInUse = reader.readElementContentAsInt("CurrentCreditInUse")
        self.currentCreditStatus = reader.readElementContentAsInt("CurrentCreditStatus")
        self.availableCredit = reader.readElementContentAsInt("AvailableCredit")
        self.amountToClear = reader.readElementContentAsInt("AmountToClear")
        self.clearanceThreshold = reader.readElementContentAsInt("ClearanceThreshold")
        self.aggregatedDebt = reader.readElementContentAsInt("AggregatedDebt")
        self.loadReferences(reader, "CreditReferences", self.creditReferences)
        self.loadReferences(reader, "ChargeReferences", self.chargeReferences)
        self.loadCreditChargeConfigurations(reader, self.creditChargeConfigurations)
        self.loadTokenGatewayConfigurations(reader, self.tokenGatewayConfigurations)
        tmp = reader.readElementContentAsString("AccountActivationTime")
        if tmp:
            self.accountActivationTime = GXDateTime(tmp)
        tmp = reader.readElementContentAsString("AccountClosureTime")
        if tmp:
            self.accountClosureTime = GXDateTime(tmp)
        self.currency.name = reader.readElementContentAsString("CurrencyName")
        self.currency.scale = reader.readElementContentAsInt("CurrencyScale")
        self.currency.unit = reader.readElementContentAsInt("CurrencyUnit")
        self.lowCreditThreshold = reader.readElementContentAsInt("LowCreditThreshold")
        self.nextCreditAvailableThreshold = reader.readElementContentAsInt("NextCreditAvailableThreshold")
        self.maxProvision = reader.readElementContentAsInt("MaxProvision")
        self.maxProvisionPeriod = reader.readElementContentAsInt("MaxProvisionPeriod")

    @classmethod
    def saveReferences(cls, writer, list_, name):
        if list_:
            writer.writeStartElement(name)
            for it in list_:
                writer.writeStartElement("Item")
                writer.writeElementString("Name", it)
                writer.writeEndElement()
            writer.writeEndElement()

    @classmethod
    def saveCreditChargeConfigurations(cls, writer, list_):
        if list_:
            writer.writeStartElement("CreditChargeConfigurations")
            for it in list_:
                writer.writeStartElement("Item")
                writer.writeElementString("Credit", it.creditReference)
                writer.writeElementString("Charge", it.chargeReference)
                writer.writeElementString("Configuration", it.collectionConfiguration)
                writer.writeEndElement()
            writer.writeEndElement()

    @classmethod
    def saveTokenGatewayConfigurations(cls, writer, list_):
        if list_:
            writer.writeStartElement("TokenGatewayConfigurations")
            for it in list_:
                writer.writeStartElement("Item")
                writer.writeElementString("Credit", it.creditReference)
                writer.writeElementString("Token", it.tokenProportion)
                writer.writeEndElement()
            writer.writeEndElement()

    def save(self, writer):
        writer.writeElementString("PaymentMode", int(self.paymentMode))
        writer.writeElementString("AccountStatus", int(self.accountStatus))
        writer.writeElementString("CurrentCreditInUse", self.currentCreditInUse)
        writer.writeElementString("CurrentCreditStatus", int(self.currentCreditStatus))
        writer.writeElementString("AvailableCredit", self.availableCredit)
        writer.writeElementString("AmountToClear", self.amountToClear)
        writer.writeElementString("ClearanceThreshold", self.clearanceThreshold)
        writer.writeElementString("AggregatedDebt", self.aggregatedDebt)
        self.saveReferences(writer, self.creditReferences, "CreditReferences")
        self.saveReferences(writer, self.chargeReferences, "ChargeReferences")
        self.saveCreditChargeConfigurations(writer, self.creditChargeConfigurations)
        self.saveTokenGatewayConfigurations(writer, self.tokenGatewayConfigurations)
        writer.writeElementString("AccountActivationTime", self.accountActivationTime)
        writer.writeElementString("AccountClosureTime", self.accountClosureTime)
        writer.writeElementString("CurrencyName", self.currency.name)
        writer.writeElementString("CurrencyScale", self.currency.scale)
        writer.writeElementString("CurrencyUnit", int(self.currency.unit))
        writer.writeElementString("LowCreditThreshold", self.lowCreditThreshold)
        writer.writeElementString("NextCreditAvailableThreshold", self.nextCreditAvailableThreshold)
        writer.writeElementString("MaxProvision", self.maxProvision)
        writer.writeElementString("MaxProvisionPeriod", self.maxProvisionPeriod)
